"""A class representing the LiveMeasurement type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class LiveMeasurement:
    """A class containing the live household electricity information."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def timestamp(self) -> str:
        """Timestamp when usage occurred"""
        return self.cache.get("timestamp")

    @property
    def power(self) -> float:
        """Consumption at the moment (Watt)"""
        return self.cache.get("power")

    @property
    def last_meter_consumption(self) -> float:
        """Last meter active import register state (kWh)"""
        return self.cache.get("lastMeterConsumption")

    @property
    def accumulated_consumption(self) -> float:
        """kWh consumed since midnight"""
        return self.cache.get("accumulatedConsumption")

    @property
    def accumulated_production(self) -> float:
        """net kWh produced since midnight"""
        return self.cache.get("accumulatedProduction")

    @property
    def accumulated_consumption_last_hour(self) -> float:
        """kWh consumed since since last hour shift"""
        return self.cache.get("accumulatedConsumptionLastHour")

    @property
    def accumulated_production_last_hour(self) -> float:
        """net kWh produced since last hour shift"""
        return self.cache.get("accumulatedProductionLastHour")

    @property
    def accumulated_cost(self) -> float:
        """Accumulated cost since midnight; requires active Tibber power deal"""
        return self.cache.get("accumulatedCost")

    @property
    def accumulated_reward(self) -> float:
        """Accumulated reward since midnight; requires active Tibber power deal"""
        return self.cache.get("accumulatedReward")

    @property
    def currency(self) -> str:
        """Currency of displayed cost; requires active Tibber power deal"""
        return self.cache.get("currency")

    @property
    def min_power(self) -> float:
        """Min consumption since midnight (Watt)"""
        return self.cache.get("minPower")

    @property
    def average_power(self) -> float:
        """Average consumption since midnight (Watt)"""
        return self.cache.get("averagePower")

    @property
    def max_power(self) -> float:
        """Peak consumption since midnight (Watt)"""
        return self.cache.get("maxPower")

    @property
    def power_production(self) -> float:
        """Net production (A-) at the moment (Watt)"""
        return self.cache.get("powerProduction")  # pragma: no cover

    @property
    def power_reactive(self) -> float:
        """Reactive consumption (Q+) at the moment (kVAr)"""
        return self.cache.get("powerReactive")  # pragma: no cover

    @property
    def power_production_reactive(self) -> float:
        """Net reactive production (Q-) at the moment (kVAr)"""
        return self.cache.get("powerProductionReactive")  # pragma: no cover

    @property
    def min_power_production(self) -> float:
        """Min net production since midnight (Watt)"""
        return self.cache.get("minPowerProduction")

    @property
    def max_power_production(self) -> float:
        """Max net production since midnight (Watt)"""
        return self.cache.get("maxPowerProduction")

    @property
    def last_meter_production(self) -> float:
        """Last meter active export register state (kWh)"""
        return self.cache.get("lastMeterProduction")

    @property
    def power_factor(self) -> float:
        """Power factor (active power / apparent power)"""
        return self.cache.get("powerFactor")  # pragma: no cover

    @property
    def voltage_phase_1(self) -> float:
        """Voltage on phase 1; on Kaifa and Aidon meters the value is not part 
        of every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("voltagePhase1")  # pragma: no cover

    @property
    def voltage_phase_2(self) -> float:
        """Voltage on phase 2; on Kaifa and Aidon meters the value is not part 
        of every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("voltagePhase2")  # pragma: no cover

    @property
    def voltage_phase_3(self) -> float:
        """Voltage on phase 3; on Kaifa and Aidon meters the value is not part 
        of every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("voltagePhase3")  # pragma: no cover

    @property
    def currentL1(self) -> float:
        """Current on L1; on Kaifa and Aidon meters the value is not part of
        every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("currentL1")  # pragma: no cover

    @property
    def currentL2(self) -> float:
        """Current on L2; on Kaifa and Aidon meters the value is not part of
        every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("currentL2")  # pragma: no cover

    @property
    def currentL3(self) -> float:
        """Current on L3; on Kaifa and Aidon meters the value is not part of
        every HAN data frame therefore the value is null at timestamps with
        second value other than 0, 10, 20, 30, 40, 50. There can be other deviations
        based on concrete meter firmware."""
        return self.cache.get("currentL3")  # pragma: no cover

    @property
    def signal_strength(self) -> int:
        """Device signal strength (Pulse - dB; Watty - percent)"""
        return self.cache.get("signalStrength")  # pragma: no cover


import asyncio
import logging
import json
from typing import Optional

import aiohttp

from tibber import API_ENDPOINT
from tibber.exceptions import APIException
from tibber.exceptions import InvalidTokenException


class QueryExecutor:
    """A class for executing sessions."""
    def __init__(self, websession: Optional[aiohttp.ClientSession] = None):
        """Instantiates the query executor. This creates a websession among other things."""
        self.logger = logging.getLogger(__name__)
        
        try:
            self.eventloop = asyncio.get_running_loop()
        except RuntimeError:
            self.logger.debug("No running event loop was found. Creating a new one with asyncio.get_event_loop()")
            self.eventloop = asyncio.get_event_loop()

        self.eventloop.run_until_complete(self.__async_init__(websession))
        
    async def __async_init__(self, websession: Optional[aiohttp.ClientSession] = None):
        if websession:
            self._websession = websession
        else:
            self.logger.debug("A websession was not provided. Creating a new aiohttp.ClientSession.")
            self._websession = aiohttp.ClientSession()

    def execute_query(self, access_token: str, query: str, retries: int = 2):
        """Executes a GraphQL query to the Tibber API.

        :param access_token: The Tibber API token to use for the request.
        :param query: The query to send to the Tibber API.
        :param retries: The amount of retries to attempt before raising an asyncio Timeout error.
        """
        post_args = self.create_request(access_token, query)
        return self.eventloop.run_until_complete(self.send_request(post_args))
        
    
    def create_request(self, access_token: str, data: str):
        """Creates a GraphQL request, but does not execute it. Returns a dict that can
        be passed to the send_request method.
        
        :param access_token: The access token to use for the request.
        :param data: The data to be sent in the request.
        :param request_type: The root type of the request (e.g. "mutation" or "query").
        """
        # TODO: Implement query variables
        payload = {"query": data, "variables": {}} 

        request = {
            "headers": {
                "Authorization": "Bearer " + access_token
            },
            "data": payload,
        }
        return request

    async def send_request(self, post_args: dict, retries: int = 2):
        """Sends a request to the Tibber API.

        :param post_args: The arguments to send in the Tibber API web request. The post args should
            contain a "headers" key with the access token authorization and a "data" key with.
        :param retries: The amount of retries to attempt before raising an asyncio Timeout error.
        """
        # TODO: Only run deepcopy if the logger level is DEBUG.
        # Log the request we are making (and redact the access token in case users share the logs)
        from copy import deepcopy
        debug_post_args = deepcopy(post_args)
        debug_post_args["headers"]["Authorization"] = "Bearer <## TOKEN REDACTED ##>"
        debug_query = debug_post_args["data"]["query"]
        debug_post_args["data"]["query"] = "<QUERY>"
        self.logger.debug("Executing a query with these post args:\n" + json.dumps(debug_post_args) + "\nWhere <QUERY> is:\n" + debug_query)

        json_response = {}
        times_attempted = 0
        while not json_response.get("data") and times_attempted <= retries:
            if times_attempted > 0:
                self.logger.info(f"Failed to retrieve data from api request. Retrying ({times_attempted} of {retries} times).")

            response = await self.websession.post(API_ENDPOINT, **post_args)
            json_response = await response.json()
            self.logger.debug("Response from API request received. The json data is:\n" + json.dumps(json_response, indent=4))

            errors = json_response.get("errors")
            if errors:
                # TODO: Handle errors better
                # For now, errors are simply logged since the method can still return data although there's an error. (see issue #6)
                self.logger.error(f"Something went wrong with the request. The following errors occured:\n{json.dumps(errors, indent=4)}")
                
                # InvalidToken:
                if "UNAUTHENTICATED" in map(lambda e: e["extensions"]["code"], errors):
                    raise InvalidTokenException("Could not authenticate with the given token!")

            times_attempted += 1

        if not json_response.get("data"):
            raise APIException(f"Something went wrong with the request. The following errors occured:\n{json.dumps(errors, indent=4)}")

        return json_response.get("data")

    @property
    def websession(self):
        return self._websession
    
    def __del__(self):
        """Close the websession when the class is deloaded"""
        if self.eventloop.is_closed(): return
        if self.websession.closed: return
        self.eventloop.run_until_complete(self.websession.close())



"""A class representing the PushNotificationResponse type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 

class PushNotificationResponse:
    def __init__(self, data: dict, tibber_client: "Account"):
        self.data = data or {}  # pragma: no cover

    @property
    def successful(self):
        return self.data.get("successful")  # pragma: no cover
    
    @property
    def pushed_to_number_of_devices(self):
        return self.data.get("pushedToNumberOfDevices")  # pragma: no cover


"""A class representing the Viewer type from the GraphQL Tibber API."""
from tibber.types.home import TibberHome
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class Viewer:
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client
        
    @property
    def name(self):
        return self.cache.get("name")

    @property
    def login(self):
        return self.cache.get("login")

    @property
    def user_id(self):
        """Unique user identifier"""
        return self.cache.get("userId")

    @property
    def account_type(self):
        """The type of account for the logged-in user."""
        return self.cache.get("accountType")
    
    @property
    def homes(self):
        """All homes visible to the logged-in user"""
        return [ TibberHome(home, self.tibber_client) for home in self.cache.get("homes", []) ]
    
    # TODO: Implement home(id: ID!): Home! method. (get_home and fetch_home)


"""A class representing the HomeFeatures type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class HomeFeatures:
    """A class to get information about the features of a TibberHome."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: Account = tibber_client

    @property
    def real_time_consumption_enabled(self) -> bool:
        """'true' if Tibber Pulse or Watty device is paired at home"""
        return self.cache.get("realTimeConsumptionEnabled")


"""A class representing the Address type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class Address:
    """An address type to get information about the location of a TibberHome."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def address1(self) -> str:
        return self.cache.get("address1")

    @property
    def address2(self) -> str:
        return self.cache.get("address2")  # pragma: no cover

    @property
    def address3(self) -> str:
        return self.cache.get("address3")  # pragma: no cover

    @property
    def city(self) -> str:
        return self.cache.get("city")  # pragma: no cover

    @property
    def postal_code(self) -> str:
        return self.cache.get("postalCode")  # pragma: no cover

    @property
    def country(self) -> str:
        return self.cache.get("country")  # pragma: no cover

    @property
    def latitude(self) -> str:
        return self.cache.get("latitude")  # pragma: no cover

    @property
    def longitude(self) -> str:
        return self.cache.get("longitude")  # pragma: no cover


from tibber.types.home_features import HomeFeatures
from tibber.types.home import NonDecoratedTibberHome
from tibber.types.home import TibberHome

from tibber.types.price_rating_threshold_percentages import PriceRatingThresholdPercentages
from tibber.types.metering_point_data import MeteringPointData
from tibber.types.price_rating_entry import PriceRatingEntry
from tibber.types.price_rating_type import PriceRatingType
from tibber.types.subscription import Subscription
from tibber.types.contact_info import ContactInfo
from tibber.types.legal_entity import LegalEntity
from tibber.types.price_rating import PriceRating
from tibber.types.price_info import PriceInfo
from tibber.types.address import Address
from tibber.types.viewer import Viewer
from tibber.types.price import Price


__version__ = "0.1.1"
DEMO_TOKEN = "5K4MVS-OjfWhK_4yrjOlFe1F6kJXPVf7eQYggo8ebAE"
API_ENDPOINT = "https://api.tibber.com/v1-beta/gql"
SUBSCRIPTION_ENDPOINT = "wss://api.tibber.com/v1-beta/gql/subscriptions"

# Import modules after defining constants to avoid circular import error.
from .account import Account

from .types.home import TibberHome
from .types.home import NonDecoratedTibberHome



"""A class representing the PriceRatingThresholdPercentages type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class PriceRatingThresholdPercentages:
    """A class to get price info."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def high(self) -> float:
        """The percentage difference when the price is considered to be 'high' (market dependent)"""
        return self.cache.get("high")

    @property
    def low(self) -> float:
        """The percentage difference when the price is considered to be 'low' (market dependent)"""
        return self.cache.get("current")


"""A class representing the Consumption type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class Consumption:
    """A class containing concrete household electricity consumption information for a time period."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def from_time(self) -> str:
        return self.cache.get("from")

    @property
    def to_time(self) -> str:
        return self.cache.get("to")

    @property
    def unit_price(self) -> float:
        return self.cache.get("unitPrice")

    @property
    def unit_price_vat(self) -> float:
        return self.cache.get("unitPriceVAT")

    @property
    def consumption(self) -> float:
        """kWh consumed"""
        return self.cache.get("consumption")

    @property
    def consumption_unit(self) -> str:
        return self.cache.get("consumptionUnit")

    @property
    def cost(self) -> float:
        return self.cache.get("cost")

    @property
    def currency(self) -> str:
        """The cost currency"""
        return self.cache.get("currency")


"""A class representing the PriceInfo type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.price import Price

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 

class PriceInfo:
    """A class to get price info."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def current(self) -> Price:
        """The energy price right now"""
        return Price(self.cache.get("current"), self.tibber_client)

    @property
    def today(self) -> list[Price]:
        """The hourly prices of the current day"""
        return [Price(hour, self.tibber_client) for hour in self.cache.get("today", [])]

    @property
    def tomorrow(self) -> list[Price]:
        """The hourly prices of the upcoming day"""
        return [Price(hour, self.tibber_client) for hour in self.cache.get("tomorrow", [])]
    
    # TODO: Implement range(resolution, first, last, ...) method


"""A class representing the Subscription type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.legal_entity import LegalEntity
from tibber.types.price_info import PriceInfo
from tibber.types.price_rating import PriceRating

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class Subscription:
    """A class to get information about the subscription of a TibberHome."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def id(self) -> str:
        return self.cache.get("id")

    @property
    def subscriber(self) -> LegalEntity:
        """The owner of the subscription"""
        return LegalEntity(self.cache.get("subscriber"), self.tibber_client)

    @property
    def valid_from(self) -> str:
        """The time the subscription started"""
        return self.cache.get("validFrom")

    @property
    def valid_to(self) -> str:
        """The time the subscription ended"""
        return self.cache.get("validTo")

    @property
    def status(self) -> str:
        """The current status of the subscription"""
        return self.cache.get("status")

    @property
    def price_info(self) -> PriceInfo:
        """Price information related to the subscription"""
        return PriceInfo(self.cache.get("priceInfo"), self.tibber_client)

    @property
    def price_rating(self) -> PriceRating:
        """Price information related to the subscription"""
        return PriceRating(self.cache.get("priceRating"), self.tibber_client)


"""Classes representing the Home type from the GraphQL Tibber API."""
import json
import asyncio
import logging
from typing import Callable
from typing import TYPE_CHECKING

import websockets

from tibber import SUBSCRIPTION_ENDPOINT
from tibber.types.legal_entity import LegalEntity
from tibber.types.address import Address
from tibber.types.metering_point_data import MeteringPointData
from tibber.types.subscription import Subscription
from tibber.types.home_features import HomeFeatures
from tibber.types.live_measurement import LiveMeasurement
from tibber.types.home_consumption_connection import HomeConsumptionConnection
from tibber.types.home_production_connection import HomeProductionConnection
from tibber.networking import QueryBuilder

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class NonDecoratedTibberHome:
    """A Tibber home with methods to get/fetch home information without the decorator functions to subscribe to live data."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client
        
        # Logging
        self.logger = logging.getLogger(__name__)

    def fetch_consumption(self, 
                          resolution: str, 
                          first: int = None, 
                          last: int = None, 
                          before: str = None, 
                          after: str = None, 
                          filter_empty_nodes: bool = False) -> HomeConsumptionConnection:
        """Consumption connection"""
        consumption_query = QueryBuilder.consumption_query(resolution, first, last, before, after, filter_empty_nodes)
        full_query = QueryBuilder.create_query("viewer", f"home(id: \"{self.id}\")", consumption_query)
        unsanitized_data = self.tibber_client.execute_query(self.tibber_client.token, full_query)

        # The format should be correct, since we requested it this way and the request
        # was successful, so we don't need to worry about key errors. 
        data = unsanitized_data["viewer"]["home"]["consumption"]
        return HomeConsumptionConnection(resolution, data, self.tibber_client)

    def fetch_production(self,
                         resolution: str = None, 
                         first: int = None, 
                         last: int = None, 
                         before: str = None, 
                         after: str = None, 
                         filter_empty_nodes: bool = False) -> HomeProductionConnection:
        production_query = QueryBuilder.production_query(resolution, first, last, before, after, filter_empty_nodes)
        full_query = QueryBuilder.create_query("viewer", f"home(id: \"{self.id}\")", production_query)
        unsanitized_data = self.tibber_client.execute_query(self.tibber_client.token, full_query)
        
        data = unsanitized_data["viewer"]["home"]["production"]
        return HomeProductionConnection(resolution, data, self.tibber_client)

    @property
    def id(self) -> str:
        return self.cache.get("id")

    @property
    def time_zone(self) -> str:
        """The time zone the home resides in"""
        return self.cache.get("timeZone")

    @property
    def app_nickname(self) -> str:
        """The nickname given to the home by the user"""
        return self.cache.get("appNickname")

    @property
    def app_avatar(self) -> str:
        """The chosen avatar for the home"""
        return self.cache.get("appAvatar")

    @property
    def size(self) -> int:
        """The size of the home in square meters"""
        return self.cache.get("size")

    @property
    def type(self) -> str:
        """The type of home."""
        return self.cache.get("type")

    @property
    def number_of_residents(self) -> int:
        """The number of people living in the home"""
        return self.cache.get("numberOfResidents")

    @property
    def primary_heating_source(self) -> str:
        """The primary form of heating in the household"""
        return self.cache.get("primaryHeatingSource")

    @property
    def has_ventilation_system(self) -> bool:
        """Whether the home has a ventilation system"""
        return self.cache.get("hasVentilationSystem")

    @property
    def main_fuse_size(self) -> int:
        """The main fuse size"""
        return self.cache.get("mainFuseSize")

    @property
    def owner(self) -> LegalEntity:
        """The registered owner of the house"""
        return LegalEntity(self.cache.get("owner"), self.tibber_client)

    @property
    def metering_point_data(self) -> MeteringPointData:
        return MeteringPointData(self.cache.get("meteringPointData"), self.tibber_client)

    @property
    def current_subscription(self) -> Subscription:
        """The current/latest subscription related to the home"""
        return Subscription(self.cache.get("currentSubscription"), self.tibber_client)
    
    @property
    def subscriptions(self) -> list:
        """All historic subscriptions related to the home"""
        return [Subscription(sub, self.tibber_client) for sub in self.cache.get("subscriptions", [])]
    
    @property
    def features(self):
        return HomeFeatures(self.cache.get("features"), self.tibber_client)
    
    # Support 1 to 1 Tibber API representation.
    @property
    def address(self) -> Address:
        return Address(self.cache.get("address"), self.tibber_client)
    
    @property
    def address1(self) -> str:
        return self.address.address1

    @property
    def address2(self) -> str:
        return self.address.address2  # pragma: no cover

    @property
    def address3(self) -> str:
        return self.address.address3  # pragma: no cover

    @property
    def city(self) -> str:
        return self.address.city  # pragma: no cover

    @property
    def postal_code(self) -> str:
        return self.address.postal_code  # pragma: no cover

    @property
    def country(self) -> str:
        return self.address.country  # pragma: no cover

    @property
    def latitude(self) -> str:
        return self.address.latitude  # pragma: no cover

    @property
    def longitude(self) -> str:
        return self.address.longitude  # pragma: no cover

class TibberHome(NonDecoratedTibberHome):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop = asyncio.get_event_loop()
        self._callbacks = {}

    def event(self, event_to_listen_for) -> Callable:
        """Returns a decorator that registers the function being
        decorated as a callback function for the given event
        
        :param event_to_listen_for: The event the decorator should register the function as a callback for.
        """
        def decorator(callback):
            """Returns the function as it is, but registers it as a callback for an event.
            
            :param callback: The function being decorated.
            :throws ValueError: if the given event is not a valid event.
            """
            if event_to_listen_for == "live_measurement":
                # Create the live_measurement key if it does not exist already
                if not ("live_measurement" in self._callbacks):
                    self._callbacks["live_measurement"] = []
                # Append the callback function to the dict of callbacks
                self._callbacks["live_measurement"].append(callback)

            else: 
                raise ValueError(f"Could not recognize the event you want to listen for: {event_to_listen_for}")
            return callback
        return decorator
        
        
    def start_livefeed(self):
        """Creates a websocket and starts pushing data out to registered callbacks."""
        self.tibber_client.eventloop.run_until_complete(self.run_websocket_loop())
        
    async def run_websocket_loop(self):
        """Starts a websocket."""
        # Tell the websocket that we are wanting to connect
        connection_init_message = json.dumps(
            {"type": "connection_init", "payload": {}}
        )
            
        # Request the data we want
        root_subscription_query = QueryBuilder.live_measurement(self.id)
        request_message = json.dumps(
            {"type": "start", "id": "1", "payload": {"query": root_subscription_query}}
        )
        
        # Connect the websocket, then fire off the created requests
        async with websockets.connect(
            SUBSCRIPTION_ENDPOINT,
            subprotocols=["graphql-ws"],
            extra_headers={"Authorization": self.tibber_client.token}
        ) as websocket:
            self.logger.debug("Sending initial payloads to start connection with live data websocket.")
            self.logger.debug("Connection init message:\n" + connection_init_message)
            self.logger.debug("Request message:\n" + request_message)
            
            await websocket.send(connection_init_message)
            await websocket.send(request_message)

            # Now we should be receiving data!
            async for data in websocket:
                dict_data = json.loads(data)
                self.process_websocket_response(dict_data)

    def process_websocket_response(self, data):
        """Processes a response with data from the live data websocket."""
        if data["type"] == "connection_ack":
            self.logger.debug("Retrieved connection_ack. The websocket connection was accepted.")
            return
        
        elif data["type"] == "ka":
            self.logger.debug("Received ka (Keep Alive) from websocket.")
            return
        
        elif data["type"] == "error":
            # TODO: Better error handling
            raise Exception("Something went wrong: " + data["payload"]["message"])
            
        # Broadcast the event
        # TODO: Differentiate between consumption data, production data and other data.
        self.broadcast_event("live_measurement", LiveMeasurement(data["payload"]["data"]["liveMeasurement"], self.tibber_client))
                    
    def broadcast_event(self, event, data):
        if not event in self._callbacks:
            self.logger.warning("The event that was broadcasted has no listeners / callbacks! Nothing was run.")
            return
        
        for callback in self._callbacks[event]:
            callback(data)


"""A class representing the PriceRating type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.price_rating_threshold_percentages import PriceRatingThresholdPercentages
from tibber.types.price_rating_type import PriceRatingType

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account 


class PriceRating:
    """A class to get the rating of a price in relative terms."""
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def threshold_percentages(self) -> dict:
        """The different 'high'/'low' price breakpoints (market dependent)"""
        return PriceRatingThresholdPercentages(self.cache.get("thresholdPercentages"), self.tibber_client)

    @property
    def hourly(self) -> dict:
        """The hourly prices of today, the previous 7 days, and tomorrow"""
        return PriceRatingType(self.cache.get("hourly"), self.tibber_client)

    @property
    def daily(self) -> dict:
        """The daily prices of today and the previous 30 days"""
        return PriceRatingType(self.cache.get("daily"), self.tibber_client)

    @property
    def monthly(self) -> dict:
        """The monthly prices of this month and the previous 31 months"""
        return PriceRatingType(self.cache.get("monthly"), self.tibber_client)